unit ExcelFormula;

{
Copyright: Algorithm (c) E. W. Bachtal, Inc

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all copies or substantial
  portions of the Software.

The software is provided "as is", without warranty of any kind, express or implied, including but not
limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In
no event shall the authors or copyright holders be liable for any claim, damages or other liability,
whether in an action of contract, tort or otherwise, arising from, out of or in connection with the
software or the use or other dealings in the software.

This is a Delphi XE port of ewbi's Excel formula parser.

Original Javascript post is: http://ewbi.blogs.com/develops/2004/12/excel_formula_p.html

Other ports are:
  C#:      http://ewbi.blogs.com/develops/2007/03/excel_formula_p.html
  Python:  http://www.ewbi.com/ewbi.develop/samples/jsport_nonEAT.html
  C++:     https://github.com/lishen2/Excel_formula_parser_cpp

Versions:
  1.0:    2011/06/13 - Orginal Delphi XE port by hadv.de
}


interface

uses
 Generics.Collections, ExcelFormulaToken, StrUtils, SysUtils;

const
 QUOTE_DOUBLE  = '"';
 QUOTE_SINGLE  = '''';
 BRACKET_CLOSE = ']';
 BRACKET_OPEN  = '[';
 BRACE_OPEN    = '{';
 BRACE_CLOSE   = '}';
 PAREN_OPEN    = '(';
 PAREN_CLOSE   = ')';
 SEMICOLON     = ';';
 WHITESPACE    = ' ';
 COMMA         = ',';
 ERROR_START   = '#';

 OPERATORS_SN = '+-';
 OPERATORS_INFIX = '+-*/^&=><';
 OPERATORS_POSTFIX = '%';

 ERRORS : array[0..6] of string = (
   '#NULL!', '#DIV/0!', '#VALUE!', '#REF!', '#NAME?', '#NUM!', '#N/A'
 );
 COMPARATORS_MULTI : array[0..2] of string = (
   '>=', '<=', '<>'
 );

type

TExcelFormula = class
 private
   FFormula: String;
   FTokens:  TList<TExcelFormulaToken>;
   //
   procedure ParseToTokens;
   function  IndexOf(const ListOfString: array of String; const Value: String): Integer;
 public
   constructor Create(Formula: String);
   destructor Destroy; override;

   function  GetCount: Integer;
   function  GetToken(Index: Integer): TExcelFormulaToken;
   function  doPop(lstStack: TStack<TExcelFormulaToken>): TExcelFormulaToken;

   property  Tokens: TList<TExcelFormulaToken> read FTokens write FTokens;


end;

implementation

uses RegularExpressions;


constructor TExcelFormula.Create(Formula: String);
begin
 if Length(Formula) = 0 then Raise EArgumentException.Create('formula');
 self.FFormula := Trim(Formula);
 FTokens := TList<TExcelFormulaToken>.Create;
 self.ParseToTokens;
end;

destructor TExcelFormula.Destroy;
var
 i: Integer;
begin
 for i := 0 to self.FTokens.Count-1 do
   self.FTokens[i].Free;
 self.FTokens.Free;
end;

function TExcelFormula.GetCount: Integer;
begin
 Result := FTokens.Count;
end;

function TExcelFormula.GetToken(Index: Integer): TExcelFormulaToken;
begin
 if (Index >= 0) and (Index < FTokens.Count) then
   Result := FTokens[Index]
 else
   Result := nil;
end;

procedure TExcelFormula.ParseToTokens;
var
 l_iIndex, l_iTokIndex: Integer;
 l_sStr, l_sValue: String;
 l_lInString, l_lInPath, l_lInRange, l_lInError, l_lIsNumber: Boolean;
 l_fVal:       Extended;
 l_oRegEx:     TRegEx;
 l_oTok, l_oToken, l_oPrevToken, l_oNextToken: TExcelFormulaToken;
 l_lstTokens:  TList<TExcelFormulaToken>;
 l_lstTokens2: TList<TExcelFormulaToken>;
 l_lstStack:   TStack<TExcelFormulaToken>;
begin
 if (Length(self.FFormula) > 2) and (self.FFormula[1] = '=') then begin
   l_lInString := False;
   l_lInPath   := False;
   l_lInRange  := False;
   l_lInError  := False;
   l_iIndex    := 2;
   l_sValue    := '';
   l_lstTokens := TList<TExcelFormulaToken>.Create;
   l_lstStack  := TStack<TExcelFormulaToken>.Create;
   l_oRegEx    := TRegEx.Create('^[1-9]{1}(\.[0-9]+)?E{1}$');
   while l_iIndex <= Length(self.FFormula) do begin
     // state-dependent character evaluation (order is important)
     // double-quoted strings
     // embeds are doubled
     // end marks token
     if l_lInString then begin
       if (self.FFormula[l_iIndex] = QUOTE_DOUBLE) then begin
         if (((l_iIndex + 2) <= Length(self.FFormula)) and (self.FFormula[l_iIndex+1] = QUOTE_DOUBLE)) then begin
           l_sValue := l_sValue + QUOTE_DOUBLE;
           Inc(l_iIndex);
         end
         else begin
           l_lInString := false;
           l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand, TExcelFormulaTokenSubtype.enText));
           l_sValue := '';
         end;
       end
       else begin
         l_sValue := l_sValue + self.FFormula[l_iIndex];
       end;
       Inc(l_iIndex);
       Continue;
      end;
      // single-quoted strings (links)
      // embeds are double
      // end does not mark a token
      if l_lInPath then begin
        if (self.FFormula[l_iIndex] = QUOTE_SINGLE) then begin
          if (((l_iIndex + 2) <= Length(self.FFormula)) and (self.FFormula[l_iIndex + 1] = QUOTE_SINGLE)) then begin
            l_sValue := l_sValue + QUOTE_SINGLE;
            Inc(l_iIndex);
          end
          else begin
            l_lInPath := false;
          end;
        end
        else begin
          l_sValue := l_sValue + self.FFormula[l_iIndex];
        end;
        Inc(l_iIndex);
        continue;
      end;
      // bracked strings (R1C1 range index or linked workbook name)
      // no embeds (changed to "()" by Excel)
      // end does not mark a token
      if l_lInRange then begin
        if (self.FFormula[l_iIndex] = BRACKET_CLOSE) then begin
          l_lInRange := false;
        end;
        l_sValue := l_sValue + self.FFormula[l_iIndex];
        Inc(l_iIndex);
        continue;
      end;
      // error values
      // end marks a token, determined from absolute list of values
      if l_lInError then begin
        l_sValue := l_sValue + self.FFormula[l_iIndex];
        Inc(l_iIndex);
        if self.IndexOf(ERRORS, UpperCase(l_sValue)) >= 0 then begin
          l_lInError := False;
          l_lstTokens.Add(TExcelFormulaToken.Create(l_sValue, TExcelFormulaTokenType.enOperand, TExcelFormulaTokenSubtype.enError));
          l_sValue := '';
        end;
        continue;
      end;
      // scientific notation check
      if Pos(self.FFormula[l_iIndex],OPERATORS_SN) > 0 then begin
        if (Length(l_sValue) > 1) then begin
          if (l_oRegEx.IsMatch(l_sValue)) then begin
            l_sValue := l_sValue + self.FFormula[l_iIndex];
            Inc(l_iIndex);
            continue;
          end;
        end;
      end;
      // independent character evaluation (order not important)
      // establish state-dependent character evaluations
      if (self.FFormula[l_iIndex] = QUOTE_DOUBLE) then begin
        if (Length(l_sValue) > 0) then begin  // unexpected
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enUnknown));
          l_sValue := '';
        end;
        l_lInString := true;
        Inc(l_iIndex);
        continue;
      end;
      if (self.FFormula[l_iIndex] = QUOTE_SINGLE) then begin
        if (Length(l_sValue) > 0) then begin // unexpected
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enUnknown));
          l_sValue := '';
        end;
        l_lInPath := true;
        Inc(l_iIndex);
        continue;
      end;
      if (self.FFormula[l_iIndex] = BRACKET_OPEN) then begin
        l_lInRange := True;
        l_sValue := l_sValue + BRACKET_OPEN;
        Inc(l_iIndex);
        continue;
      end;
      if (self.FFormula[l_iIndex] = ERROR_START) then begin
        if (Length(l_sValue) > 0) then begin // unexpected
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enUnknown));
          l_sValue := '';
        end;
        l_lInError := true;
        l_sValue := l_sValue + ERROR_START;
        Inc(l_iIndex);
        continue;
      end;
      // mark start and end of arrays and array rows
      if (self.FFormula[l_iIndex] = BRACE_OPEN) then begin
        if (Length(l_sValue) > 0) then begin // unexpected
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enUnknown));
          l_sValue := '';
        end;
        l_oTok := TExcelFormulaToken.create('ARRAY', TExcelFormulaTokenType.enFunction, TExcelFormulaTokenSubtype.enStart);
        l_lstTokens.Add(l_oTok);
        l_lstStack.Push(l_oTok);
        l_oTok := TExcelFormulaToken.create('ARRAYROW', TExcelFormulaTokenType.enFunction, TExcelFormulaTokenSubtype.enStart);
        l_lstTokens.Add(l_oTok);
        l_lstStack.Push(l_oTok);
        Inc(l_iIndex);
        continue;
      end;
      if (self.FFormula[l_iIndex] = SEMICOLON) then begin
        if (Length(l_sValue) > 0) then begin
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
          l_sValue := '';
        end;
        l_lstTokens.Add(self.doPop(l_lstStack));
        l_lstTokens.Add(TExcelFormulaToken.create(',', TExcelFormulaTokenType.enArgument));
        l_oTok := TExcelFormulaToken.create('ARRAYROW', TExcelFormulaTokenType.enFunction, TExcelFormulaTokenSubtype.enStart);
        l_lstTokens.Add(l_oTok);
        l_lstStack.Push(l_oTok);
        Inc(l_iIndex);
        continue;
      end;
      if (self.FFormula[l_iIndex] = BRACE_CLOSE) then begin
        if (Length(l_sValue) > 0) then begin
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
          l_sValue := '';
        end;
        l_lstTokens.Add(self.doPop(l_lstStack));
        l_lstTokens.Add(self.doPop(l_lstStack));
        Inc(l_iIndex);
        continue;
      end;
      // trim white-space
      if (self.FFormula[l_iIndex] = WHITESPACE) then begin
        if (Length(l_sValue) > 0) then begin
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
          l_sValue := '';
        end;
        l_lstTokens.Add(TExcelFormulaToken.create('', TExcelFormulaTokenType.enWhitespace));
        Inc(l_iIndex);
        while ((self.FFormula[l_iIndex] = WHITESPACE) and (l_iIndex < Length(self.FFormula))) do begin
          Inc(l_iIndex);
        end;
        continue;
      end;
      // multi-character comparators
      if ((l_iIndex + 2) <= Length(self.FFormula)) then begin
        if (self.IndexOf(COMPARATORS_MULTI, Copy(self.FFormula,l_iIndex,2)) <> -1) then begin
          if (Length(l_sValue) > 0) then begin
            l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
            l_sValue := '';
          end;
          l_lstTokens.Add(TExcelFormulaToken.create(Copy(self.FFormula,l_iIndex,2), TExcelFormulaTokenType.enOperatorInfix, TExcelFormulaTokenSubtype.enLogical));
          l_iIndex := l_iIndex + 2;
          continue;
        end;
      end;
      // standard infix operators
      if Pos(self.FFormula[l_iIndex],OPERATORS_INFIX) > 0 then begin
        if (Length(l_sValue) > 0) then begin
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
          l_sValue := '';
        end;
        l_lstTokens.Add(TExcelFormulaToken.create(self.FFormula[l_iIndex], TExcelFormulaTokenType.enOperatorInfix));
        Inc(l_iIndex);
        continue;
      end;
      // standard postfix operators (only one)
      if Pos(self.FFormula[l_iIndex],OPERATORS_POSTFIX) > 0 then begin
        if (Length(l_sValue) > 0) then begin
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
          l_sValue := '';
        end;
        l_lstTokens.Add(TExcelFormulaToken.create(self.FFormula[l_iIndex], TExcelFormulaTokenType.enOperatorPostfix));
        Inc(l_iIndex);
        continue;
      end;
      // start subexpression or function
      if (self.FFormula[l_iIndex] = PAREN_OPEN) then begin
        if (Length(l_sValue) > 0) then begin
          l_oTok := TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enFunction, TExcelFormulaTokenSubtype.enStart);
          l_lstTokens.Add(l_oTok);
          l_lstStack.Push(l_oTok);
          l_sValue := '';
        end
        else begin
          l_oTok := TExcelFormulaToken.create('', TExcelFormulaTokenType.enSubexpression, TExcelFormulaTokenSubtype.enStart);
          l_lstTokens.Add(l_oTok);
          l_lstStack.Push(l_oTok);
        end;
        Inc(l_iIndex);
        continue;
      end;
      // function, subexpression, or array parameters, or operand unions
      if (self.FFormula[l_iIndex] = COMMA) then begin
        if (Length(l_sValue) > 0) then begin
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
          l_sValue := '';
        end;
        if l_lstStack.Count > 0 then begin
          if TExcelFormulaToken(l_lstStack.Peek).TokType = TExcelFormulaTokenType.enFunction then begin
            l_lstTokens.Add(TExcelFormulaToken.create(',', TExcelFormulaTokenType.enArgument));
          end
          else begin
            l_lstTokens.Add(TExcelFormulaToken.create(',', TExcelFormulaTokenType.enOperatorInfix, TExcelFormulaTokenSubtype.enUnion));
          end;
        end;
        Inc(l_iIndex);
        continue;
      end;
      // stop subexpression
      if (self.FFormula[l_iIndex] = PAREN_CLOSE) then begin
        if (Length(l_sValue) > 0) then begin
          l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
          l_sValue := '';
        end;
        l_lstTokens.Add(self.doPop(l_lstStack));
        Inc(l_iIndex);
        continue;
      end;
      //
      // token accumulation
      l_sValue := l_sValue + self.FFormula[l_iIndex];
      Inc(l_iIndex);
    end;
    //
    // dump remaining accumulation
    if (Length(l_sValue) > 0) then begin
      l_lstTokens.Add(TExcelFormulaToken.create(l_sValue, TExcelFormulaTokenType.enOperand));
    end;
    //
    // move tokenList to new set, excluding unnecessary white-space tokens and converting necessary ones to intersections
    l_lstTokens2 := TList<TExcelFormulaToken>.Create;
    l_iTokIndex := 0;
    while (l_iTokIndex < l_lstTokens.Count) do begin
      l_oToken := l_lstTokens.Items[l_iTokIndex];
      if l_oToken = nil then begin
        Inc(l_iTokIndex);
        Continue;
      end;
      if (l_oToken.TokType <> TExcelFormulaTokenType.enWhitespace) then begin
        l_lstTokens2.Add(l_oToken);
        Inc(l_iTokIndex);
        Continue;
      end;
      if (l_iTokIndex = 0) or (l_iTokIndex = (l_lstTokens.Count-1)) then begin
        Inc(l_iTokIndex);
        continue;
      end;
      l_oPrevToken := nil;
      l_oNextToken := nil;
      if l_iTokIndex > 0 then
        l_oPrevToken := l_lstTokens.Items[l_iTokIndex-1];
      if l_iTokIndex < (l_lstTokens.Count-1) then
        l_oNextToken := l_lstTokens.Items[l_iTokIndex+1];
      if (l_oPrevToken = nil) then begin
        Inc(l_iTokIndex);
        continue;
      end;
      if not(
           ((l_oPrevToken.TokType = TExcelFormulaTokenType.enFunction) and (l_oPrevToken.TokSubType = TExcelFormulaTokenSubtype.enStop))
            or
           ((l_oPrevToken.TokType = TExcelFormulaTokenType.enSubexpression) and (l_oPrevToken.TokSubType = TExcelFormulaTokenSubtype.enStop))
            or
            (l_oPrevToken.TokType = TExcelFormulaTokenType.enOperand) ) then begin
        Inc(l_iTokIndex);
        continue;
      end;
      if l_oNextToken = nil then begin
        Inc(l_iTokIndex);
        continue;
      end;
      if not(
            ((l_oNextToken.TokType = TExcelFormulaTokenType.enFunction) and (l_oNextToken.TokSubType = TExcelFormulaTokenSubtype.enStart))
            or
            ((l_oNextToken.TokType = TExcelFormulaTokenType.enSubexpression) and (l_oNextToken.TokSubType = TExcelFormulaTokenSubtype.enStart))
            or
             (l_oNextToken.TokType = TExcelFormulaTokenType.enOperand)
            ) then begin
        Inc(l_iTokIndex);
        continue;
      end;
      l_lstTokens2.Add(TExcelFormulaToken.create('', TExcelFormulaTokenType.enOperatorInfix, TExcelFormulaTokenSubtype.enIntersection));
      Inc(l_iTokIndex);
    end;
 		l_lstTokens.Clear;   // Reset without Items destroy
    // move tokens to final list, switching infix "-" operators to prefix when appropriate, switching infix "+" operators
    // to noop when appropriate, identifying operand and infix-operator subtypes, and pulling "@" from function names
    l_iTokIndex := 0;
    while (l_iTokIndex < l_lstTokens2.Count) do begin
      l_oToken := l_lstTokens2.Items[l_iTokIndex];
      if (l_oToken = nil) then begin
        Inc(l_iTokIndex);
        continue;
      end;
      l_oPrevToken := nil;
      l_oNextToken := nil;
      if l_iTokIndex > 0 then
        l_oPrevToken := l_lstTokens2.Items[l_iTokIndex-1];
      if l_iTokIndex < (l_lstTokens2.Count-1) then
        l_oNextToken := l_lstTokens2.Items[l_iTokIndex+1];
      if ((l_oToken.TokType = TExcelFormulaTokenType.enOperatorInfix) and (l_oToken.TokValue = '-')) then begin
        if (l_iTokIndex = 0) then begin
          l_oToken.TokType := TExcelFormulaTokenType.enOperatorPrefix;
        end
        else if
            ((l_oPrevToken.TokType = TExcelFormulaTokenType.enFunction) and (l_oPrevToken.TokSubType = TExcelFormulaTokenSubtype.enStop))
            or
            ((l_oPrevToken.TokType = TExcelFormulaTokenType.enSubexpression) and (l_oPrevToken.TokSubType = TExcelFormulaTokenSubtype.enStop))
            or
            (l_oPrevToken.TokType = TExcelFormulaTokenType.enOperatorPostfix)
            or
            (l_oPrevToken.TokType = TExcelFormulaTokenType.enOperand) then begin
          l_oToken.TokSubType := TExcelFormulaTokenSubtype.enMath;
        end
        else begin
          l_oToken.TokType := TExcelFormulaTokenType.enOperatorPrefix;
        end;
        l_lstTokens.Add(l_oToken);
        Inc(l_iTokIndex);
        continue;
      end;
      if ((l_oToken.TokType = TExcelFormulaTokenType.enOperatorInfix) and (l_oToken.TokValue = '+')) then begin
        if l_iTokIndex = 0 then begin
          Inc(l_iTokIndex);
          continue;
        end
        else if (
             ((l_oPrevToken.TokType = TExcelFormulaTokenType.enFunction) and (l_oPrevToken.TokSubType = TExcelFormulaTokenSubtype.enStop))
              or
             ((l_oPrevToken.TokType = TExcelFormulaTokenType.enSubexpression) and (l_oPrevToken.TokSubType = TExcelFormulaTokenSubtype.enStop))
              or
             (l_oPrevToken.TokType  = TExcelFormulaTokenType.enOperatorPostfix)
              or
             (l_oPrevToken.TokType  = TExcelFormulaTokenType.enOperand) ) then begin
           l_oToken.TokSubtype := TExcelFormulaTokenSubtype.enMath;
        end
        else begin
          Inc(l_iTokIndex);
          continue;
        end;
        l_lstTokens.Add(l_oToken);
        Inc(l_iTokIndex);
        continue;
      end;
      if ((l_oToken.TokType = TExcelFormulaTokenType.enOperatorInfix) and
            (l_oToken.TokSubtype = TExcelFormulaTokenSubtype.enNothing)) then begin
        if Pos(l_oToken.TokValue[1],'<>=') > 0 then
          l_oToken.TokSubType := TExcelFormulaTokenSubtype.enLogical
        else if (l_oToken.TokValue = '&') then
          l_oToken.TokSubType := TExcelFormulaTokenSubtype.enConcatenation
        else
          l_oToken.TokSubType := TExcelFormulaTokenSubtype.enMath;
        l_lstTokens.Add(l_oToken);
        Inc(l_iTokIndex);
        continue;
      end;
      if ((l_oToken.TokType = TExcelFormulaTokenType.enOperand) and (l_oToken.TokSubtype = TExcelFormulaTokenSubtype.enNothing)) then begin
        l_sStr :=  StrUtils.ReplaceText(l_oToken.TokValue, '.', '');
        l_sStr :=  StrUtils.ReplaceText(l_sStr, ',', '.');       // DecimalSeparator := '.';
        try
          l_fVal := StrToFloat(l_sStr);
          l_lIsNumber := True;
        except
          l_lIsNumber := False;
        end;
        if (not l_lIsNumber) then begin
          if (UpperCase(l_oToken.TokValue) = 'TRUE') or (UpperCase(l_oToken.TokValue) = 'FALSE') then
            l_oToken.TokSubtype := TExcelFormulaTokenSubtype.enLogical
          else
            l_oToken.TokSubtype := TExcelFormulaTokenSubtype.enRange;
        end
        else
          l_oToken.TokSubtype := TExcelFormulaTokenSubtype.enNumber;
        l_lstTokens.Add(l_oToken);
        Inc(l_iTokIndex);
        continue;
      end;
      if (l_oToken.TokType = TExcelFormulaTokenType.enFunction) then begin
        if (Length(l_oToken.TokValue) > 0) then begin
          if (l_oToken.TokValue[1] = '@') then begin
            l_oToken.TokValue := Copy(l_oToken.TokValue,2,999);
          end;
        end;
      end;
      l_lstTokens.Add(l_oToken);
      Inc(l_iTokIndex);
    end;
    self.FTokens := l_lstTokens;
    l_lstTokens2.Clear;
    l_lstStack.Clear;
    l_lstTokens2.Free;
    l_lstStack.Free;
  end;
end;

function TExcelFormula.IndexOf(const ListOfString: array of String; const Value: String): Integer;
var
 i: Integer;
begin
 Result := -1;
 for i := Low(ListOfString) to High(ListOfString) do begin
   if ListOfString[i] = Value then begin
     Result := i;
     break;
   end;
 end;
end;

function TExcelFormula.doPop(lstStack: TStack<TExcelFormulaToken>): TExcelFormulaToken;
var
 l_oToken: TExcelFormulaToken;
begin
 if (lstStack.Count = 0) then
   Result := nil
 else begin
   l_oToken := lstStack.Pop();
   Result :=  TExcelFormulaToken.create('', l_oToken.TokType, TExcelFormulaTokenSubtype.enStop);
 end;
end;

end.
